package cz.zcu.kiv.vsp.msloup.statistic;

import cz.zcu.fav.kiv.jsim.JSimHead;
import cz.zcu.fav.kiv.jsim.JSimProcess;
import cz.zcu.kiv.vsp.msloup.request.IRequest;

/**
 * Trida pocita zakladni statistiku, tj.:
 * <ul>
 * <li>stredni delku fronty (Lw)</li>
 * <li>prumernou stredni dobu odezvy serveru (Tq)</li>
 * <li>soucet strednich dob odezvy serveru</li>
 * <li>zatizeni serveru</li>
 * <li>pocet zpracovanych pozadavku</li>
 * </ul>
 * Pouziti:<br>
 * <pre><code> Room r = new Room(...);
 * IStatistics statistics = new BasicStatistics();
 * r.setStatistics(statistics);
 * ...
 * </code></pre>
 * @author Martin Sloup
 *
 */
public class BasicStatistics implements IStatistics {
	protected JSimProcess source;
	protected JSimHead queue;

	private long requestCount;
	private double sumOfTq;
	private double sumOfProcessingTime;

	/**
	 * Konstruktor tridy
	 */
	public BasicStatistics() {
		reset();
	}

	/**
	 * Nastvi pocatecni hodnoty pocitanych statistik
	 */
	public void reset() {
		requestCount = 0;
		sumOfTq = 0;
		sumOfProcessingTime = 0;
	}

	@Override
	public void initialize(JSimProcess source, JSimHead queue) {
		this.source = source;
		this.queue = queue;

		reset();
	}

	@Override
	public void processRequest(IRequest request, double queueEnterTime,
			double processingTime) {

		requestCount++;

		sumOfTq+= source.getParent().getCurrentTime() - queueEnterTime;
		sumOfProcessingTime+=processingTime;


	}

	/**
	 * Vraci stredni delku fronty zdroje
	 * @return	stredni delka fronty
	 */
	public double getLw() {
		return queue.getLw();
	}

	/**
	 * Vraci prumernou stredni dobu odezvy serveru
	 * @return	prumernou stredni dobu odezvy serveru
	 */
	public double getTq() {
		return sumOfTq / requestCount;
	}

	/**
	 * Vraci soucet strednich dob odezvy serveru za vsechny pozadavky
	 * @return	soucet stredni doby odezvy
	 */
	public double getSumOfTq() {
		return sumOfTq;
	}

	/**
	 * Vraci zatizeni serveru
	 * @return	zatizeni serveru
	 */
	public double getLoad() {
		return sumOfProcessingTime / source.getParent().getCurrentTime();
	}

	/**
	 * Vraci pocet zpracovanych pozadavku
	 * @return	pocet zpracovanych pozadavku
	 */
	public long getRequestCount() {
		return requestCount;
	}

	/**
	 * Vraci referenci na objekt zdroje, na kterem se pocita statistika
	 * @return	referenci na zdroj
	 */
	protected JSimProcess getSource() {
		return source;
	}

	/**
	 * Vraci frontu zdroje
	 * @return	fronta zdroje
	 */
	protected JSimHead getQueue() {
		return queue;
	}

	/**
	 * Vraci stredni pocet pozadavku zdroje
	 * @return	stredni pocet zdroje
	 */
	public double getLq() {
		return getLoad() / (1 - getLoad());
	}

}
