package cz.zcu.kiv.vsp.msloup;

import java.text.DecimalFormat;
import java.text.NumberFormat;

import cz.zcu.fav.kiv.jsim.JSimInvalidParametersException;
import cz.zcu.fav.kiv.jsim.JSimMethodNotSupportedException;
import cz.zcu.fav.kiv.jsim.JSimSecurityException;
import cz.zcu.fav.kiv.jsim.JSimSimulation;
import cz.zcu.fav.kiv.jsim.JSimSimulationAlreadyTerminatedException;
import cz.zcu.fav.kiv.jsim.JSimTooManyHeadsException;
import cz.zcu.fav.kiv.jsim.JSimTooManyProcessesException;
import cz.zcu.kiv.vsp.msloup.generator.ExponentialGenerator;
import cz.zcu.kiv.vsp.msloup.generator.GaussianGenerator;
import cz.zcu.kiv.vsp.msloup.object.Choice;
import cz.zcu.kiv.vsp.msloup.object.Room;
import cz.zcu.kiv.vsp.msloup.object.Terminator;
import cz.zcu.kiv.vsp.msloup.patient.PatientGenerator;
import cz.zcu.kiv.vsp.msloup.statistic.BasicStatistics;
import cz.zcu.kiv.vsp.msloup.statistic.QueueStatistics;
import cz.zcu.kiv.vsp.msloup.util.Counter;

/**
 * Trida spoustejici simulaci
 * @author Martin Sloup
 *
 */
public class Main {
	/**
	 * Vstupni bod hlavniho programu
	 * @param args	argumenty programu
	 */
	public static void main(String[] args) {
		if (args.length == 2) {
			long steps = Config.DEFAULT_STEPS;
			try { steps = Integer.parseInt(args[0]); } catch (Exception e) {}
			if (args[1].equalsIgnoreCase("EXP")) {
				runSimulation(steps);
			} else if (args[1].equalsIgnoreCase("GAUSS")) {
				runSimulation(Config.DEFAULT_STEPS, Config.C1);
				runSimulation(Config.DEFAULT_STEPS, Config.C2);
				runSimulation(Config.DEFAULT_STEPS, Config.C3);
			}
		} else {
			runSimulation(Config.DEFAULT_STEPS);
			runSimulation(Config.DEFAULT_STEPS, Config.C1);
			runSimulation(Config.DEFAULT_STEPS, Config.C2);
			runSimulation(Config.DEFAULT_STEPS, Config.C3);
		}
	}

	/**
	 * Spusti simulaci v exponencialnim rozdeleni
	 * @param requests		pocet generovanychg pozadavku
	 */
	public static void runSimulation(long requests) {
		runSimulation(requests, null);
	}

	/**
	 * Spsuti simulaci v normalovem (Gauss) rozdeleni
	 * @param requests		pocet pozadavku
	 * @param c				koeficient C=sigma/mi
	 */
	public static void runSimulation(long requests, Double c) {
		try {
			JSimSimulation simulation = new JSimSimulation("Simulace pacientu");

			//v simulaci jsou 4 mistnosti
			Room[] room = new Room[4];

			//dva generatory pacienty
			PatientGenerator[] generator = new PatientGenerator[2];

			//citace vygenerovanych a prijatych pacientu
			Counter counterInput = new Counter(requests);
			Counter counterOutput = new Counter(requests);

			//je-li koeficient null, pak bude exponencialni doba obsluhy u mistnosti a generatoru
			if (c == null) {
				simulation.message("\nSpoustim simulaci pro exponencialni rozdeleni:");
				room[0] = new Room("Obvodni lekar", simulation, new ExponentialGenerator(Config.MI1));
				room[1] = new Room("Chirurgie", simulation, new ExponentialGenerator(Config.MI2));
				room[2] = new Room("Sadrovna", simulation, new ExponentialGenerator(Config.MI3));
				room[3] = new Room("Rentgen", simulation, new ExponentialGenerator(Config.MI4));

				generator[0] = new PatientGenerator("Poliklinika", simulation, new ExponentialGenerator(Config.GENERATOR_LAMBDA1), counterInput);
				generator[1] = new PatientGenerator("Nemocnice", simulation, new ExponentialGenerator(Config.GENERATOR_LAMBDA2), counterInput);;
			} else {
				simulation.message("\nSpoustim simulaci pro normalove (Gauss) rozdeleni:");
				simulation.message("Koeficient C=" + c);

				//gaussovska doba obsluhy mistnosti
				room[0] = new Room("Obvodni lekar", simulation, new GaussianGenerator(1./Config.MI1, c * Config.MI1));
				room[1] = new Room("Chirurgie", simulation, new GaussianGenerator(1./Config.MI2, c * Config.MI2));
				room[2] = new Room("Sadrovna", simulation, new GaussianGenerator(1./Config.MI3, c * Config.MI3));
				room[3] = new Room("Rentgen", simulation, new GaussianGenerator(1./Config.MI4, c * Config.MI4));

				generator[0] = new PatientGenerator("Poliklinika", simulation, new GaussianGenerator(Config.GENERATOR_MI1, c * Config.GENERATOR_MI1), counterInput);
				generator[1] = new PatientGenerator("Nemocnice", simulation, new GaussianGenerator(Config.GENERATOR_MI2, c * Config.GENERATOR_MI2), counterInput);;
			}

			//dva ukoncovaci prvky
			Terminator[] terminator = new Terminator[2];
			terminator[0] = new Terminator(counterOutput, simulation);
			terminator[1] = new Terminator(counterOutput, simulation);

			//dve vetveni
			Choice[] choice = new Choice[2];

			//nastaveni vetveni a propojeni mistnosti
			generator[0].setOutput(room[0]);
			choice[0] = new Choice();
			choice[0].addChoice(Config.P1, terminator[0]);
			choice[0].addChoice(room[1]);

			room[0].setOutput(choice[0]);

			generator[1].setOutput(room[1]);

			choice[1] = new Choice();
			choice[1].addChoice(Config.P2, room[2]);
			choice[1].addChoice(Config.P3, room[3]);
			choice[1].addChoice(terminator[1]);

			room[1].setOutput(choice[1]);

			room[2].setOutput(room[1]);
			room[3].setOutput(room[1]);

			//pridani statistiky do mistnosti
			room[0].setStatistics(new BasicStatistics());
			room[1].setStatistics(new QueueStatistics());
			room[2].setStatistics(new BasicStatistics());
			room[3].setStatistics(new BasicStatistics());

			//Spusteni generatoru pacientu
			for(PatientGenerator g : generator)
				g.activateNow();

			//spusteni simulace
			while(counterOutput.getValue() > 0) simulation.step();

			// ukonceni simulace
			simulation.shutdown();

			//vypsani statistik
			long totalRequest = 0;
			double lq = 0;
			simulation.message("Simulace prerusena v case " + simulation.getCurrentTime());
			for (PatientGenerator item : generator) {
				simulation.message(item.getName() + " - generovano pacientu: " + item.getRequestCount());
				totalRequest += item.getRequestCount();

			}

			long sumOfRequest = 0;
			for (Room item : room) {
				simulation.message("Statistika mistosti " + item.getName() + ":");

				BasicStatistics statistics = (BasicStatistics) item.getStatistics();
				simulation.message("  Lq = " + statistics.getLq());
				simulation.message("  Tq = " + statistics.getTq());
				simulation.message("  load = " + statistics.getLoad());
				simulation.message("  Lw = " + statistics.getLw());

				sumOfRequest+= statistics.getRequestCount();
				lq+= statistics.getLq();
			}

			sumOfRequest = 0;
			double sumOfTq = 0;
			for(Terminator t: terminator) {
				sumOfRequest+= t.getRequestCount();
				sumOfTq+=t.getSumOfTq();
			}

			simulation.message("Celkova statistika:");
			simulation.message(" Lq = " + lq);
			simulation.message(" Tq = " + sumOfTq/sumOfRequest);

			simulation.message("Celkovy pocet pacientu: " + totalRequest);

			QueueStatistics statistics = (QueueStatistics) room[1].getStatistics();
			simulation.message("Statistika fronty " + room[1].getName() + ":");
			simulation.message("  Stredni delka: " + statistics.getQueueLengthMedian());
			simulation.message("  Rozptyl: " + statistics.getQueueLengthVariance());
			simulation.message("  Smerodatna odchylka: " + statistics.getQueueLengthSigma());
			simulation.message("  Minimalni delka: " + statistics.getMinQueueLength());
			simulation.message("  maximalni delka: " + statistics.getMaxQueueLength());
			simulation.message("Histogram:");
			simulation.message(getGraphicHistogram(statistics));


		} catch (JSimInvalidParametersException e) {
			e.printStackTrace();
		} catch (JSimTooManyHeadsException e) {
			e.printStackTrace();
		} catch (JSimSimulationAlreadyTerminatedException e) {
			e.printStackTrace();
		} catch (JSimTooManyProcessesException e) {
			e.printStackTrace();
		} catch (JSimMethodNotSupportedException e) {
			e.printStackTrace();
		} catch (JSimSecurityException e) {
			e.printStackTrace();
		}


	}

	/**
	 * Vraci graficke vyjadreni histogramu fronty
	 * @param statistics	objekt statistiky fronty
	 * @return	graficke vyjadreni histogramu
	 */
	private static String getGraphicHistogram(QueueStatistics statistics) {
		StringBuilder sb = new StringBuilder();
		NumberFormat numberFormat = new DecimalFormat("00000");
		NumberFormat doubleFormat = new DecimalFormat("0000.0000");

		int size = Config.HISTOGRAM_COLS;
		int[] histogram = statistics.getQueueLengthHistogram(size);
		double step = statistics.getQueueLengthHistogramStep(size);

		int max_value = 0;
		for(int i = 0; i < histogram.length; i++) {
			if (histogram[i] > max_value) max_value = histogram[i];
		}

		for(int i = 0; i < histogram.length; i++) {
			long value = histogram[i];
			sb.append(doubleFormat.format(statistics.getMinQueueLength() + i * step));
			sb.append(". - ");
			sb.append(numberFormat.format(value));
			sb.append(" - ");

			int count = (int) (((double)value / max_value) * Config.MAX_STAR_LENGTH);

			for(int y = 0; y < count; y++)
				sb.append(Config.STAR_CHAR);

			sb.append('\n');
		}

		return sb.toString();
	}
}
