package cz.zcu.kiv.vsp.msloup.object;

import cz.zcu.fav.kiv.jsim.JSimHead;
import cz.zcu.fav.kiv.jsim.JSimInvalidParametersException;
import cz.zcu.fav.kiv.jsim.JSimLink;
import cz.zcu.fav.kiv.jsim.JSimProcess;
import cz.zcu.fav.kiv.jsim.JSimProcessDeath;
import cz.zcu.fav.kiv.jsim.JSimSecurityException;
import cz.zcu.fav.kiv.jsim.JSimSimulation;
import cz.zcu.fav.kiv.jsim.JSimSimulationAlreadyTerminatedException;
import cz.zcu.fav.kiv.jsim.JSimTooManyHeadsException;
import cz.zcu.fav.kiv.jsim.JSimTooManyProcessesException;
import cz.zcu.kiv.vsp.msloup.generator.IGenerator;
import cz.zcu.kiv.vsp.msloup.request.IRequest;
import cz.zcu.kiv.vsp.msloup.request.IRequestInput;
import cz.zcu.kiv.vsp.msloup.statistic.IStatistics;

/**
 * Predstavuje mistnost lekare
 * @author Martin Sloup
 */
public class Room extends JSimProcess implements IRequestInput {
	private JSimHead queue;
	private IGenerator generator;
	private IRequestInput output;
	private IStatistics statistics;

	/**
	 * Konstruktor serveru.
	 * @param name		nazev serveru
	 * @param simulation	objekt simulatoru
	 * @param generator		objekt generatoru doby obsluhy pacienta
	 * @exception JSimSimulationAlreadyTerminatedException
	 *                Vyjimka je vyhozena, pokud simulace byla jiz ukocena.
	 * @exception JSimInvalidParametersException
	 *                Vyjimka je vyhozena, pokud neni specifikovan objekt simulatoru.
	 * @throws JSimTooManyHeadsException
	 *                Vyjimka je vyhozena, pokud neni mozne jiz do simulatoru pridat dalsi proces.
	 * @exception JSimTooManyProcessesException
	 *                Vyjimka je vyhozena, pokud neni mozne jiz do simulatoru pridat dalsi proces.
	 */
	public Room(String name, JSimSimulation simulation, IGenerator generator)
			throws JSimInvalidParametersException, JSimTooManyHeadsException, JSimSimulationAlreadyTerminatedException, JSimTooManyProcessesException {
		super(name, simulation);

		this.generator = generator;

		this.queue = new JSimHead(name + " - Queue", simulation);
	}

	/**
	 * Prijme pozadavek v simulaci
	 * @param request	pozadavek v simulaci
	 */
	@Override
	public void receiveRequest(IRequest request) {
		JSimLink link = new JSimLink(request);
		try {
			link.into(queue);
			if (isIdle())
				activate(myParent.getCurrentTime());
		} catch (JSimSecurityException e) {
			System.err.println("Nepovedlo se pridat pozadavek do fronty. J-Sim hlasi: " + e.getMessage());
		} catch (JSimInvalidParametersException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@Override
	protected void life() {
		try {
			while(true) {
				//pokud je prazdna fronta uspime process
				if (queue.empty()) {
					passivate();
				} else {
					double processingTime = Math.abs(generator.nextValue());

					JSimLink link = queue.first();

					IRequest request = (IRequest) link.getData();

					hold(processingTime);

					if (statistics != null)
						statistics.processRequest(request, link.getEnterTime(), processingTime);

					link.out();

					if (output != null)
						output.receiveRequest(request);

				}
			}
		} catch (JSimSecurityException e) {
			e.printStackTrace();
		} catch (JSimProcessDeath e) {
			//konec
			//e.printStackTrace();
		} catch (JSimInvalidParametersException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Nastavuje prijemce pacienta po z pracovani
	 * @param output	prijemce pacienta
	 */
	public void setOutput(IRequestInput output) {
		this.output = output;
	}

	/**
	 * Vraci objekt, ktery prijima pacienty po zpracovani
	 * @return	objekt prijemce pacienta
	 */
	public IRequestInput getOutput() {
		return output;
	}

	/**
	 * Vraci generator doby obsluhy pacienta
	 * @return	generator doby obsluhy
	 */
	public IGenerator getGenerator() {
		return generator;
	}

	/**
	 * Nastavuje instanci tridy, ktera zpracovava statistiku
	 * @param statistic	instance tridy na spracovani statistiky
	 */
	public void setStatistics(IStatistics statistic) {
		this.statistics = statistic;
		this.statistics.initialize(this, queue);
	}

	/**
	 * Vraci instanci tridy, ktera spracovava statistiku
	 * @return	instance tridy na spracovani statistiky
	 */
	public IStatistics getStatistics() {
		return statistics;
	}
}
