package cz.zcu.kiv.vsp.msloup.statistic;

import java.util.HashMap;

import cz.zcu.kiv.vsp.msloup.request.IRequest;

/**
 * Pocita statistiku fronty. Krome zakladnich statistik z BasicStatistics
 * pocita:
 * <ul>
 * <li>stredni delku fronty</li>
 * <li>rozptyl delky fronty</li>
 * <li>Sigmu pro delku fronty</li>
 * <li>maximalni delku fronty</li>
 * <li>minimalni delku fronty</li>
 * <li>histogram</li>
 * </ul>
 * Pouziti:<br>
 * <pre><code> Room r = new Room(...);
 * IStatistics statistics = new QueueStatistics();
 * r.setStatistics(statistics);
 * ...
 * </code></pre>
 * @author Martin Sloup
 *
 */
public class QueueStatistics extends BasicStatistics {
	private long sumOfQueueLength;
	private long sumOfQueueLengthSquare;
	private long minQueueLength;
	private long maxQueueLength;
	private HashMap<Long, Long> queueHistogramMap;

	@Override
	public void reset() {
		super.reset();

		sumOfQueueLength = 0;
		sumOfQueueLengthSquare = 0;
		minQueueLength = Long.MAX_VALUE;
		maxQueueLength = Long.MIN_VALUE;
		queueHistogramMap = new HashMap<Long, Long>();
	}

	@Override
	public void processRequest(IRequest request, double queueEnterTime,
			double processingTime) {
		super.processRequest(request, queueEnterTime, processingTime);

		// odeceteme aktualni pozadavek od celkoveho poctu pozadavku fronty
		long queueLength = queue.cardinal() - 1;

		sumOfQueueLength+= queueLength;
		sumOfQueueLengthSquare+= queueLength * queueLength;

		//nastavime minumum / maximum delky fronty
		if (queueLength > maxQueueLength)
			maxQueueLength = queueLength;
		if (queueLength < minQueueLength)
			minQueueLength = queueLength;

		//pridame do histogramu delku fronty
		if (!queueHistogramMap.containsKey(queueLength)) {
			queueHistogramMap.put(queueLength, 1L);
		} else {
			queueHistogramMap.put(queueLength, queueHistogramMap.get(queueLength) + 1L);
		}
	}

	/**
	 * Vrac stedn delku fronty
	 * @return	stedn delka fronty
	 */
	public double getQueueLengthMedian() {
		return ((double)sumOfQueueLength) / getRequestCount();
	}

	/**
	 * Vrac rozptyl delky fronty
	 * @return	rozptyl delky fronty
	 */
	public double getQueueLengthVariance() {
		double median = getQueueLengthMedian();

		return (((double)sumOfQueueLengthSquare) / getRequestCount()) - median * median;
	}

	/**
	 * Vraci smerodatnou odchylku delky fronty
	 * @return	smerodatnou odchylku delky fronty
	 */
	public double getQueueLengthSigma() {
		return Math.sqrt(getQueueLengthVariance());
	}

	/**
	 * Vraci maximalni delku fronty
	 * @return	maximalni delka fronty
	 */
	public long getMaxQueueLength() {
		return maxQueueLength;
	}

	/**
	 * Vraci minimalni delku fronty
	 * @return	minimalni delka fronty
	 */
	public long getMinQueueLength() {
		return minQueueLength;
	}

	/**
	 * Vraci velikost kroku histogramu fronty
	 * @param size	velikost histrogramu (pocet sloupcu)
	 * @return	velikost kroku
	 */
	public double getQueueLengthHistogramStep(int size) {
		return Math.abs((double)maxQueueLength-minQueueLength) / size;
	}

	/**
	 * Vraci histogram fronty
	 * @param size pocet sloupcu histogramu
	 * @return	histogram fronty
	 */
	public int[] getQueueLengthHistogram(int size) {
		int[] histogram = new int[size + 1];
		double step = getQueueLengthHistogramStep(size);

		for(Long index: queueHistogramMap.keySet()) {
			Long value = queueHistogramMap.get(index);
			//pridani hodnoty do histogramu
			int hIndex = (int) (((double)(index - minQueueLength)) / step);
			histogram[hIndex]+=value;
		}
		return histogram;
	}
}
